from marker.converters.pdf import PdfConverter
from marker.models import create_model_dict
from marker.output import text_from_rendered
from broai.experiments.utils import experiment

@experiment
def pdf_to_markdown(pdf_path: str) -> str:

    converter = PdfConverter(
        artifact_dict=create_model_dict(),
    )
    rendered = converter(pdf_path)
    markdown_text, _, images = text_from_rendered(rendered)
    return markdown_text, images


from datetime import datetime
from zoneinfo import ZoneInfo  # Available in Python 3.9+

def get_timestamp():
    dt = datetime.now(ZoneInfo("Asia/Bangkok"))
    return dt.strftime('%Y-%m-%d %H:%M:%S.%f')

from rich import print

def success(msg: str):
    print(f"[bold green][✔] {msg}[/bold green]")

def error(msg: str):
    print(f"[bold red][✖] {msg}[/bold red]")



from typing import Any, List, Optional, Union
from pydantic import BaseModel, Field
from broai.prompt_management.utils import get_input, get_return_json_schema

class Persona(BaseModel):
    name: str = Field(description="An agent's name.")
    description: str = Field(description="A short description describes an agent's characteristics.")

    def as_prompt(self):
        return f"Persona:\n\t- Name: {self.name}\n\t- Description: {self.description}"

class Instructions(BaseModel):
    instructions:List[str] = Field(description="The instructions (tasks) that the agent must do, make it as imperative sentence.")
    cautions:Optional[List[str]] = Field(description="Cautions is unlike tasks, but if you don't do this, something bad or error might happen.", default=None)

    def as_prompt(self):
        instructions_str = "Instructions:\n"+"\n".join([f"\t- {text}" for text in self.instructions])
        cautions = self.cautions
        if cautions is not None and isinstance(cautions, list):
            cautions_str = "\nCautions:\n"+"\n".join([f"\t- {text}" for text in cautions])
            return instructions_str+cautions_str
        return instructions_str
    
class Example(BaseModel):
    setting: str = Field(description="The setting of the example")
    input: Union[str, BaseModel] = Field(description="The input example based on InputFormat of each agent.")
    output: Union[str, BaseModel] = Field(description="The output example based on OutputFormat of each agent.")

class Examples(BaseModel):
    examples:List[Example] = Field(description="List of examples to be followed.")

    def _validate_input(self, obj):
        if isinstance(obj, str):
            return obj
        if isinstance(obj, BaseModel):
            return get_input(obj)
        raise TypeError(f"{type(obj)} is not implemented yet, try {type(str)} or {type(BaseModel)}")

    def _validate_output(self, obj):
        if isinstance(obj, str):
            return obj
        if isinstance(obj, BaseModel):
            return get_return_json_schema(obj)
        raise TypeError(f"{type(obj)} is not implemented yet, try {type(str)} or {type(BaseModel)}")
    
    def as_prompt(self):
        examples = []
        for enum, example in enumerate(self.examples):
            ex_input = self._validate_input(example.input)
            ex_output = self._validate_output(example.output)
            prompt = f"""Example {enum+1}: "{example.setting}"\n{ex_input}\n\n{ex_output}"""
            examples.append(prompt)
        return "\n\n".join(examples)


        


import click
from pathlib import Path
import shutil
from broai.utils import success, error
import pyfiglet
from rich import print
from broai.__version__ import __version__
from broai.cli.version_bump import bump_version

def show_banner():
    ascii_banner = pyfiglet.figlet_format("BroAI")
    print(f"[bold cyan]{ascii_banner}[/bold cyan]")
    print("[cyan]🚀 Welcome to BroAI CLI[/cyan]")
    print("[cyan]Usage:[/cyan] broai [command]")
    print("[cyan]Example:[/cyan] broai init")

@click.group(invoke_without_command=True)
@click.version_option(version=__version__)
@click.pass_context
def main(ctx):
    """BroAI CLI Tool."""
    if ctx.invoked_subcommand is None:
        show_banner()

def jupyter_init():
    """Real logic to initialize the project (can be called from Python or CLI)."""
    agents_dir = Path.cwd() / "agents"

    if agents_dir.exists():
        error("Folder 'agents/' already exists.")
        return

    agents_dir.mkdir(parents=True)
    success("Created folder 'agents/'")

    (agents_dir / "__init__.py").write_text("# Agents package\n")
    success("Created file 'agents/__init__.py'")

    template_src = Path(__file__).parent / "agent_template.py"
    template_dst = agents_dir / "agent_template.py"

    if template_src.exists():
        shutil.copy(template_src, template_dst)
        success("Copied 'agent_template.py' to 'agents/agent_template.py'")
    else:
        error("WARNING: Template file not found. Skipped copying.")

@main.command()
def init():
    """Initialize a new BroAI project."""
    jupyter_init()

@main.command()
@click.argument("path", type=click.Path(file_okay=False, dir_okay=True))  # Remove exists=True
@click.argument("agent_name", type=str)
def add_agent(path, agent_name):
    """Create a new agent Python file at the specified location."""
    # Define the source template file
    template_src = Path(__file__).parent / "agent_template.py"
    
    # Define the destination file path
    agent_file = Path(path) / f"{agent_name}.py"

    # Create the directory if it doesn't exist
    agent_file.parent.mkdir(parents=True, exist_ok=True)  # Will create the full path if necessary

    # Check if the agent file already exists
    if agent_file.exists():
        print(f"[bold red]❌ File '{agent_file}' already exists![/bold red]")
        return

    # Check if the template exists
    if not template_src.exists():
        print(f"[bold red]❌ Template file '{template_src}' not found![/bold red]")
        return
    
    # Copy the template to the new agent file
    shutil.copy(template_src, agent_file)

    print(f"[bold green]✅ Created agent file at '{agent_file}' from template.[/bold green]")


@main.command(hidden=True)
@click.argument("part", type=click.Choice(["patch", "minor", "major"]))
@click.option("--auto", is_flag=True, help="Automatically bump based on the latest PyPI version")
def bump(part, auto):
    """Bump the version (patch, minor, major) and sync with PyPI."""
    if auto:
        package_name = "broai"  # Update this if needed
        bump_version(part, package_name)
    else:
        bump_version(part, "broai")



from broai.prompt_management.core import PromptGenerator
from broai.prompt_management.interface import Persona, Instructions, Examples, Example
from broai.experiments.bro_agent import BroAgent
from pydantic import BaseModel, Field
from typing import Tuple, List, Dict, Any, Optional

# you can use any model sharing the same methods: .run, .SystemMessage, .UserMessage, .AIMessage
from broai.llm_management.ollama import BedrockOllamaChat
bedrock_model = BedrockOllamaChat(
    model_name="us.meta.llama3-2-1b-instruct-v1:0",
    temperature=0,
)

class InputMessage(BaseModel):
    message:str = Field(description="A user's input message that you have to understand and respond according to the task")

class StructuredOutput(BaseModel):
    respond:str = Field(description="An agent's respond to the message based on the task provided in the instructions")

prompt_generator = PromptGenerator(
    persona=Persona(name="John", description="a helpful assistant"),
    instructions=Instructions(
        instructions=[
            "something John must do to complete the task"
        ],
        cautions=[
            "something John must not do. if he does, he may fail the task"
        ]
    ),
    structured_output=StructuredOutput,
    examples=Examples(examples=[
        Example(
            setting="A setting of an example, i.e. a journal article, a blog post, a social encounter",
            input=InputMessage(message="This is a user's input message"),
            output=StructuredOutput(respond="This is what an agent should respond")
        )
    ]),
    fallback=StructuredOutput(respond="Fall back is used when parsing structured_output fails or LLM service yields error")
)

agent = BroAgent(
    prompt_generator=prompt_generator,
    model=bedrock_model
)


from broai.duckdb_management.utils import get_create_table_query, get_insert_query
from broai.interface import Context
from typing import List, Dict, Any, Literal, Tuple
import duckdb
import os
import json
from broai.experiments.utils import experiment
from broai.experiments.huggingface_embedding import BaseEmbeddingModel
from functools import wraps

def validate_baseclass(input_instance, input_name, baseclass):
    if not isinstance(input_instance, baseclass):
        raise TypeError(f"{input_name} must be of type, {baseclass.__name__}. Instead got {type(input_instance)}")
    return input_instance

def get_json_where_query(field:str, filter_metadata:Dict[str, Any]=None):
    if filter_metadata is None:
        return ""
    query = []
    str_int = lambda k, v: f"{field} ->> '{k}' = '{v}'" if isinstance(v, str) else f"CAST({field} ->> '{k}' AS INTEGER) = {v}"
    for k, v in filter_metadata.items():
        query.append(str_int(k, v))
    return f"WHERE {' AND '.join(query)}"

class BaseVectorStore:
    def __init__(self, db_name:str, table, schemas:Dict[str, Any]):
        self.db_name = db_name
        self.table = table
        self.__schemas = schemas
        self.create_table()

    def sql(self, query, params:Dict[str, Any]=None):
        with duckdb.connect(self.db_name) as con:
            con.sql(query, params=params)

    def sql_df(self, query, params:Dict[str, Any]=None):
        with duckdb.connect(self.db_name) as con:
            df = con.sql(query, params=params).to_df()
        return df

    def executemany(self, query, rows:List[Tuple[Any]]):
        with duckdb.connect(self.db_name) as con:
            con.executemany(query, rows)
    
    def show_schemas(self):
        return self.__schemas

    def delete_table(self):
        query = f"DELETE FROM {self.table};"
        self.sql(query)

    def drop_table(self)->None:
        query = f"""DROP TABLE {self.table};"""
        self.sql(query)

    def create_table(self,):
        query = get_create_table_query(table=self.table, schemas=self.show_schemas())
        self.sql(query)

    def remove_database(self, confirm:str=None)->None:
        if confirm == f"remove {self.db_name}":
            os.remove(self.db_name)
            return
        print(f"If you want to remove database, use confirm 'remove {self.db_name}'")

@experiment
class DuckVectorStore(BaseVectorStore):
    """
    A vector store backed by DuckDB.

    Args:
        db_name (str): Path to the DuckDB file (e.g., './duckmemory.db').
        table (str): Name of the table to store embeddings.
        embedding (BaseEmbeddingModel): An embedding model that implements the `.run()` method.
        limit (int, optional): Default number of top results to return. Defaults to 5.
    """
    def __init__(self, db_name:str, table:str, embedding:BaseEmbeddingModel, limit:int=5):
        self.embedding_model = validate_baseclass(embedding, "embedding", BaseEmbeddingModel)
        self.embedding_size = self.embedding_model.run(["test"]).shape[1]
        self.limit = limit
        schemas = {
            "id": "VARCHAR",
            "context": "VARCHAR",
            "metadata": "JSON",
            "type": "VARCHAR",
            "embedding": f"FLOAT[{self.embedding_size}]",
            "created_at": "TIMESTAMP",
            "updated_at": "TIMESTAMP",
        }
        super().__init__(db_name=db_name, table=table, schemas=schemas)

    @staticmethod
    def with_fts_index(func):
        @wraps(func)
        def wrapper(self, *args, **kwargs):
            result = func(self, *args, **kwargs)
            self.create_fts_index()
            return result
        return wrapper
    
    # ---- [Create] ---- #
    @with_fts_index
    def add_contexts(self, contexts:List[Context]):
        id_list = [c.id for c in contexts]
        context_list = [c.context for c in contexts]
        metadata_list = [c.metadata for c in contexts]
        type_list = [c.type for c in contexts]
        embedding_list = self.embedding_model.run(context_list)
        created_list = [c.created_at for c in contexts]
        rows = list(zip(id_list, context_list, metadata_list, type_list, embedding_list, created_list, created_list))
        query = f"INSERT INTO {self.table} (id, context, metadata, type, embedding, created_at, updated_at) VALUES (?, ?, ?, ?, ?, ?, ?)"
        self.executemany(query, rows)

    # ---- [Read] ---- #
    def read_all(self):
        query = f"SELECT * FROM {self.table};"
        return self.sql_df(query)

    def read_by_ids(self, ids:List[str]):
        params = ", ".join([f'{i}' for i in ids])
        query = f"SELECT * FROM {self.table} WHERE id IN ({params});"
        return self.sql_df(query)

    def vector_search(self, search_query:str, filter_metadata:Dict[str, Any]=None, context:bool=True, limit:int=None):
        if limit is None or not isinstance(limit, int):
            limit = self.limit
        vector = self.embedding_model.run(sentences=[search_query])[0]
        filter_query = get_json_where_query(field="metadata", filter_metadata=filter_metadata)
        query = f"""SELECT *, array_cosine_similarity(embedding, $searchVector::FLOAT[{self.embedding_size}]) AS score FROM {self.table} {filter_query} ORDER BY score DESC LIMIT {limit};"""
        if context is False:
            return self.sql_df(query=query, params=dict(searchVector=vector))
        return self.sql_contexts(query=query, params=dict(searchVector=vector))

    def fulltext_search(self, search_query:str, limit:int=5, context:bool=True):
        query = f"""\
        SELECT *
        FROM (
            SELECT *, fts_main_{self.table}.match_bm25(
                id,
                '{search_query}',
                fields := 'context'
            ) AS score
            FROM {self.table}
        ) sq
        ORDER BY score DESC;
        """
        if context is False:
            return self.sql_df(query=query, params=None)
        return self.sql_contexts(query=query, params=None)

    # ---- [Update] ---- #
    @with_fts_index
    def update_contexts(self, contexts:List[Context]):
        id_list = [c.id for c in contexts]
        context_list = [c.context for c in contexts]
        metadata_list = [c.metadata for c in contexts]
        created_list = [c.created_at for c in contexts]
        rows = list(zip(context_list, metadata_list, created_list, id_list))
        query = f"UPDATE {self.table} SET context = ?, metadata = ?, updated_at = ? WHERE id = ?"
        self.executemany(query, rows)

    # ---- [Delete] ---- #
    @with_fts_index
    def delete_contexts(self, contexts:List[Context]):
        rows = [(c.id,) for c in contexts]
        query = f"DELETE FROM {self.table} WHERE id = ?"
        self.executemany(query, rows)

    # ---- [Utils] ---- #
    def sql_contexts(self, query, params:Dict[str, Any]=None):
        df = self.sql_df(query, params=params)
        if df.shape[0]==0:
            return
        df = df.loc[~df['score'].isna(), ["id", "context", "metadata", "type", "updated_at"]].copy()
        return [Context(id=record["id"], context=record["context"], metadata=json.loads(record["metadata"]), type=record['type'], created_at=record['updated_at']) for record in df.to_dict(orient="records")]

    def create_fts_index(self):
        query = f"""
        INSTALL fts;
        LOAD fts;
        PRAGMA create_fts_index(
            '{self.table}', 'id', 'context', overwrite=1
        );
        """.strip()
        self.sql(query)



import requests
import re
from pathlib import Path
from rich import print
import tomlkit

def get_latest_pypi_version(package_name: str) -> str:
    """Fetch the latest version of the package from PyPI."""
    url = f"https://pypi.org/pypi/{package_name}/json"
    response = requests.get(url)
    
    if response.status_code != 200:
        print(f"[bold red]❌ Failed to fetch the latest version from PyPI![/bold red]")
        return None
    
    data = response.json()
    return data["info"]["version"]

def bump_version(part: str, package_name: str):
    """Bump the version (patch, minor, major) in __version__.py and pyproject.toml."""
    latest_version = get_latest_pypi_version(package_name)
    
    if not latest_version:
        return

    # Extract the version parts from the latest PyPI version
    major, minor, patch = map(int, latest_version.split('.'))

    if part == "patch":
        patch += 1
    elif part == "minor":
        minor += 1
        patch = 0
    elif part == "major":
        major += 1
        minor = 0
        patch = 0

    new_version = f'{major}.{minor}.{patch}'
    print(f"[bold blue]Latest version on PyPI: {latest_version}[/bold blue]")
    print(f"[bold green]Bumping to new version: {new_version}[/bold green]")

    # Now bump the version in __version__.py and pyproject.toml
    update_version_files(new_version)

def update_version_files(new_version: str):
    """Update both __version__.py and pyproject.toml with the new version."""
    version_file = Path(__file__).parent.parent / "__version__.py"
    version_file.write_text(f'__version__ = "{new_version}"\n')

    print(f"[bold green]✅ Bumped version in __version__.py to {new_version}[/bold green]")

    # Update pyproject.toml
    toml_file = Path(__file__).parent.parent.parent / "pyproject.toml"
    with open(toml_file, "r") as f:
        toml_data = tomlkit.load(f)
    
    toml_data["project"]["version"] = new_version

    with open(toml_file, "w") as f:
        tomlkit.dump(toml_data, f)

    print(f"[bold green]✅ Bumped version in pyproject.toml to {new_version}[/bold green]")


from FlagEmbedding import BGEM3FlagModel
from typing import Literal, List, Optional, Any, Protocol
from pydantic import BaseModel, Field
from broai.experiments.utils import experiment
from enum import Enum
import numpy as np
from abc import ABC, abstractmethod

class EmbeddingDimension(Enum):
    BAAI_BGE_M3 = 1024

class BaseEmbeddingModel(ABC):
    @abstractmethod
    def run(self, sentences: List[str]) -> np.ndarray:
        ...

@experiment
class BAAIEmbedding(BaseEmbeddingModel, BaseModel):
    model_id: Literal["BAAI/bge-m3"] = Field(default="BAAI/bge-m3")
    use_fp16: bool = Field(default=True)
    batch_size:Optional[int] = Field(default=None)
    max_length:Optional[int] = Field(default=None)
    model: Optional[Any] = None

    def __init__(self, **data):
        super().__init__(**data)
        self.model = self.get_model()

    def get_model(self):
        params = {}
        params['model_name_or_path'] = self.model_id
        params['use_fp16'] = self.use_fp16
        if self.batch_size:
            params['batch_size'] = self.batch_size
        if self.max_length:
            params['max_length'] = self.max_length
        return BGEM3FlagModel(**params)
    
    def run(self, sentences: List[str]):
        if not isinstance(sentences, list):
            raise TypeError(f"sentences must be of type list with str as element, not {type(sentences)}")
        vectors = self.model.encode(sentences)
        return vectors['dense_vecs']


import enum
from typing import Any, Dict, List, Literal, Optional, Union
from pydantic import BaseModel, Field, field_validator
from uuid import uuid4
from datetime import datetime
from broai.utils import get_timestamp
from enum import Enum

class Context(BaseModel):
    id:str = Field(description="unique id generated by system", default_factory=lambda: str(uuid4()))
    context:str = Field(description="context or chunk of information that will be stored and used in RAG system") 
    metadata:Optional[Dict[str, Any]] = Field(description="", default=None)
    type:Literal["document"] = Field(description="type of ducument", default="document")
    created_at: datetime = Field(description="Timestamp Asia/Bangkok", default_factory=get_timestamp)

class Contexts(BaseModel):
    contexts:List[Context] = Field(default_factory=list)

    def add_context(self, context:Context):
        if not isinstance(context, Context):
            raise TypeError(f"context must be of type: Context, not {type(context)}.")
        self.contexts.append(context)

    def add_contexts(self, contexts:List[Context]):
        if not isinstance(contexts, list):
            raise TypeError(f"contexts must be of type list, not {type(contexts)}")
        for context in contexts:
            self.add_context(context=context)

    def as_knowledge(self):
        contexts = []
        for context in self.contexts:
            metadata = context.metadata.copy()
            c = context.context
            contexts.append(
                f"Source: {metadata.get('source', 'unknown')}\nContext: \n{c}"
            )
        prompt = '\n\n'.join(contexts)
        prompt = f"Knowledge: \n\n{prompt}"
        return prompt

class TaskStatus(str, Enum):
    NOT_STARTED = "not_started"
    PENDING = "pending"
    PROCESSING = "processing"
    DONE = "done"
    ERROR = "error"
    CANCELED = "canceled"           # Optional: User/system canceled
    TIMEOUT = "timeout"             # Optional: Process timed out
    RETRYING = "retrying"           # Optional: Retry attempt in progress


from typing import Any, Callable, List, Union
from pydantic import BaseModel
from broai.prompt_management.core import PromptGenerator
from broai.prompt_management.utils import get_input, get_json_schema, parse_json_output
from broai.llm_management.interface import LLMChatInterface

class BroAgent:
    def __init__(
            self,
            prompt_generator:PromptGenerator,
            model:LLMChatInterface,
            retry:int=3,
            tools:List[Callable] = None
    ):
        self.prompt_generator = prompt_generator
        self.model = model
        self.retry = retry
        self.tools = tools
        self.cnt = 0
        self.__errors = []

    def get_error_msg(self, errors:List[str])->str:
        if len(errors)==0:
            return ""
        return "Avoid the following errors: \n\n{error_msg}".format(error_msg="\n".join(errors))

    def response_pydantic(self, text:str):
        try:
            parsed_json = parse_json_output(text)
            structure_output = self.prompt_generator.structured_output.model_validate_json(parsed_json)
            return structure_output
        except Exception as e:
            self.__errors.append(str(e))
            self.cnt+=1
            return None
    
    def _run(self, request):
        response = self.model.run(
            system_prompt=self.prompt_generator.as_prompt(), 
            messages=[self.model.UserMessage(request + "\n" + self.get_error_msg(self.__errors))]
        )
        if isinstance(self.prompt_generator.structured_output, str) or self.prompt_generator.structured_output is None:
            return response
        return self.response_pydantic(response)

    def read_errors(self):
        return self.__errors
    
    def run(self, request:Union[str, BaseModel])->Any:
        self.cnt=0
        if isinstance(request, BaseModel):
            request = get_input(request)
        while self.cnt < self.retry:
            response = self._run(request)
            if response is not None:
                self.cnt=0
                return response
        if response is None:
            # fallback is None -> default message
            if self.prompt_generator.fallback is None:
                return "unknown error"
            # in other case return as it is
            return self.prompt_generator.fallback



